// dasciicam.c (c) 2009 Dan Stowell, all rights reserved.
// Published under the GNU GPL version 3.
// Contains some fragments of code from hasciicam (big up hasciicam folks!)
// but mainly coded from scratch using the V4L2 docs (big up V4L2 docs folks!)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <signal.h>

#include <linux/videodev2.h>

#include <aalib.h>

static const char *version = "0.2";

int fd;  //filedescriptor for video device
int buftype = V4L2_BUF_TYPE_VIDEO_CAPTURE;

unsigned char *grey;
int YtoRGB[256];

aa_context *ascii_context;
struct aa_renderparams *ascii_rndparms;
struct aa_hardware_params ascii_hwparms;
struct aa_savedata ascii_save;

// sig handler:
void quitproc (int Sig);
volatile sig_atomic_t userbreak;

/*
void YUV422_to_grey(unsigned char *src, unsigned char *dst, int w, int h) {
  int c,cc, size;
  size=w*h;
  for (c=0,cc=0;c<size;++c,cc+=2)
    dst[c] = YtoRGB[src[cc]];
}
*/

int xstep=2, ystep=4;
int xbytestep;
int ybytestep;
int renderhop=2, framenum=0; // renderhop is how many frames to guzzle before rendering
int gw, gh; // number of cols/rows in grey intermediate representation
int vbytesperline;
/*
 * slooooooooooow method but easy to understand:
void YUV422_to_grey(unsigned char *src, unsigned char *dst, int w, int h) {
  int writepos=0, x, y;
  for (y=0; y<h; y += ystep)
    for (x=0; x<w; x += xstep)
      dst[writepos++] = YtoRGB[src[(y*w + x) * 2]];
}
*/
void YUV422_to_grey(unsigned char *src, unsigned char *dst, int w, int h) {
  unsigned char *writehead, *readhead;
  int x,y;
  writehead = dst;
  readhead  = src;
  for(y=0; y<gh; ++y){
    for(x=0; x<gw; ++x){
      *(writehead++) = *readhead;
      readhead += xbytestep;
    }
    readhead += ybytestep;
  }
}


int main(int argc, char **argv){

  // Get options
  int c;
  //char *cvalue = NULL;
  while ((c = getopt(argc, argv, "w:t:r:Vh")) != -1){
    switch (c){
      case 'w': /* width */
        xstep = atoi(optarg);
        break;
      case 't': /* heighT */
        ystep = atoi(optarg);
        break;
      case 'r': /* rate */
        renderhop = atoi(optarg);
        break;
      case 'V': /* version */
        printf("dasciicam %s\n", version);
        printf("Published under the terms of the GPLv3\n");
        exit(0);
      case 'h': /* help */
        printf("dasciicam, (c) Dan Stowell, published under the GPLv3+\n\n"
"Options:\n"
"  -w [num]     width divisor (e.g. 2 to shrink width by 2)\n"
"  -t [num]     height divisor (e.g. 2 to shrink height by 2)\n"
"  -r [num]     rate divisor (e.g. 2 to update half as often as source)\n"
"  -V           print version number\n"
"  -h           show this help");
        //cvalue = optarg;
	exit(0);
    }
  }

  const char *devfile = "/dev/video0";
  int errno;
  if (-1 == (fd = open(devfile,O_RDWR|O_NONBLOCK))) {
    perror("!! error in opening video capture device: ");
    return -1;
  } else {
    close(fd);
    fd = open(devfile,O_RDWR);
  }


// Check that streaming is supported
struct v4l2_capability capability;
memset(&capability, 0, sizeof(capability));
if(-1 == ioctl(fd, VIDIOC_QUERYCAP, &capability)) {
	perror("VIDIOC_QUERYCAP");
	exit(EXIT_FAILURE);
}
if((capability.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0){
	printf("Fatal: Device %s does not support video capture\n", capability.card);
	exit(EXIT_FAILURE);
}
if((capability.capabilities & V4L2_CAP_STREAMING) == 0){
	printf("Fatal: Device %s does not support streaming data capture\n", capability.card);
	exit(EXIT_FAILURE);
}

// Switch to the first video input (example 1-2)
int index=0;
if(-1 == ioctl(fd, VIDIOC_S_INPUT, &index)) {
	perror("VIDIOC_S_INPUT");
	exit(EXIT_FAILURE);
}
printf("Switched to first input\n");

// Get info about current video input (example 1-1)
struct v4l2_input input;
memset(&input, 0, sizeof(input));
input.index = index;
if(-1 == ioctl(fd, VIDIOC_ENUMINPUT, &input)) {
	perror("VIDIOC_ENUMINPUT");
	exit(EXIT_FAILURE);
}
printf("Current input: %s supports:\n", input.name);
// example 1-6
struct v4l2_standard standard;
memset(&standard, 0, sizeof(standard));
standard.index = 0;
while(0 == ioctl (fd, VIDIOC_ENUMSTD, &standard)){
	if(standard.id & input.std)
		printf("   - %s\n", standard.name);
	standard.index++;
}

// Need to find out (request?) specific data format (sec 1.10.1)
struct v4l2_format format;
memset(&format, 0, sizeof(format));
format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//noeffect   format.fmt.pix.width  = 320;
//noeffect   format.fmt.pix.height = 240;
if(-1 == ioctl(fd, VIDIOC_G_FMT, &format)) {
	perror("VIDIOC_G_FMT");
	exit(EXIT_FAILURE);
}
// the format.fmt.pix member, a v4l2_pix_format, is now filled in
printf("Current capture is %u x %u, format %4.4s, bytes-per-line %u\n", 
	format.fmt.pix.width, format.fmt.pix.height, (char*)&format.fmt.pix.pixelformat, format.fmt.pix.bytesperline);

// OK, we've found the default, now let's request greyscale
//format.fmt.pix.pixelformat = 0x47824559; // 'GREY'
format.fmt.pix.pixelformat = 0x56595559; // 'YUYV', write it backwards
//format.fmt.pix.pixelformat = 0x59458247; // 'GREY', write it backwards
if(-1 == ioctl(fd, VIDIOC_S_FMT, &format)) {
	perror("VIDIOC_S_FMT");
	exit(EXIT_FAILURE);
}
if(-1 == ioctl(fd, VIDIOC_G_FMT, &format)) {
	perror("VIDIOC_G_FMT");
	exit(EXIT_FAILURE);
}
// the format.fmt.pix member, a v4l2_pix_format, is now filled in
printf("Current capture is %u x %u, format %4.4s, bytes-per-line %u\n", 
	format.fmt.pix.width, format.fmt.pix.height, (char*)&format.fmt.pix.pixelformat, format.fmt.pix.bytesperline);

int vw = format.fmt.pix.width;
int vh = format.fmt.pix.height;
vbytesperline = format.fmt.pix.bytesperline;
xbytestep = xstep + xstep; // for YUV422. for other formats may differ
ybytestep = vbytesperline * (ystep-1);
// we shrink our pixels crudely, by hopping over them:
gw = vw / xstep;
gh = vh / ystep;
// aalib converts every block of 4 pixels to one character, so our sizes shrink by 2:
int aw = gw / 2;
int ah = gh / 2;

size_t greysize = gw * gh;
grey = malloc(greysize); // To get grey from YUYV we simply ignore the U and V bytes
printf("grey buffer is %i bytes in size\n", greysize);
int j;
for (j=0; j< 256; ++j)
  YtoRGB[j] = 1.164*(j-256);


  /* initialize aalib default params */
  memcpy (&ascii_hwparms, &aa_defparams, sizeof (struct aa_hardware_params));
  ascii_rndparms = aa_getrenderparams();
  /* width/height image setup */
  ascii_hwparms.font = NULL; // default font, thanks
  ascii_hwparms.width = aw;
  ascii_hwparms.height = ah;
  /* gathering aalib commandline options */
  aa_parseoptions (&ascii_hwparms, ascii_rndparms, &argc, argv);


//if (mode > 0)
//    ascii_context = aa_init (&save_d, &ascii_hwparms, &ascii_save);
//  else
    ascii_context = aa_autoinit (&ascii_hwparms);

  if(!ascii_context) {
    fprintf(stderr,"!! cannot initialize aalib\n");
    exit(-1);
  }
ascii_rndparms->bright   = 60;
ascii_rndparms->contrast = 4;
ascii_rndparms->gamma    = 3;









struct v4l2_requestbuffers reqbuf;
struct {
        void *start;
        size_t length;
} *buffers;
unsigned int i;

memset (&reqbuf, 0, sizeof (reqbuf));
reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
reqbuf.memory = V4L2_MEMORY_MMAP;
reqbuf.count = 32;

if (-1 == ioctl (fd, VIDIOC_REQBUFS, &reqbuf)) {
        if (errno == EINVAL)
                printf ("Fatal: Video capturing by mmap-streaming is not supported\n");
        else
                perror ("VIDIOC_REQBUFS");

        exit (EXIT_FAILURE);
}

/* We want at least five buffers. */
/* ...or do we not care...?
if (reqbuf.count < 5) {
        printf ("Not enough buffer memory\n");
        exit (EXIT_FAILURE);
}
*/

printf("Happily we have access to %i buffers\n", reqbuf.count);

buffers = calloc (reqbuf.count, sizeof (*buffers));
if (buffers == NULL){
     printf("calloc failure!");
     exit (EXIT_FAILURE);
}

struct v4l2_buffer buffer;
for (i = 0; i < reqbuf.count; i++) {

        memset (&buffer, 0, sizeof (buffer));
        buffer.type = reqbuf.type;
	buffer.memory = V4L2_MEMORY_MMAP;
        buffer.index = i;

        if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buffer)) {
                perror ("VIDIOC_QUERYBUF");
                exit (EXIT_FAILURE);
        }

        buffers[i].length = buffer.length; /* remember for munmap() */

        buffers[i].start = mmap (NULL, buffer.length,
                                 PROT_READ | PROT_WRITE, /* recommended */
                                 MAP_SHARED,             /* recommended */
                                 fd, buffer.m.offset);

        if (MAP_FAILED == buffers[i].start) {
                /* If you do not exit here you should unmap() and free()
 *                    the buffers mapped so far. */
                perror ("mmap");
                exit (EXIT_FAILURE);
        }
}



// OK, the memory is mapped, so let's queue up the buffers, turn on streaming, and do the business

// queue up all the buffers for the first time
for (i = 0; i < reqbuf.count; i++) {

        memset (&buffer, 0, sizeof (buffer));
        buffer.type = reqbuf.type;
	buffer.memory = V4L2_MEMORY_MMAP;
        buffer.index = i;

        if (-1 == ioctl (fd, VIDIOC_QBUF, &buffer)) {
                perror ("VIDIOC_QBUF");
                exit (EXIT_FAILURE);
        }	
}

// turn on streaming
if(-1 == ioctl(fd, VIDIOC_STREAMON, &buftype)) {
	perror("VIDIOC_STREAMON");
	exit(EXIT_FAILURE);
}


for (i = 0; i < greysize; i++) {
	grey[i] = i % 160; //256;
}


// now the big runloop of queuing and dequeuing, asciifying, etc
while(1){	
//for(i = 0; i < 1000; ++i){
	// Can we have a buffer please?
        if (-1 == ioctl (fd, VIDIOC_DQBUF, &buffer)) {
                perror ("VIDIOC_DQBUF");
                exit (EXIT_FAILURE);
        }	

	if((++framenum) == renderhop){
		framenum=0;
		YUV422_to_grey(buffers[buffer.index].start, grey, vw, vh);

		memcpy( aa_image(ascii_context), grey, greysize);
		aa_fastrender(ascii_context, 0, 0, vw/(xstep*2), vh/(ystep*2)); //TODO are the w&h args correct?
//		aa_render(ascii_context, ascii_rndparms, 0, 0, vw/(xstep*2), vh/(ystep*2)); //TODO are the w&h args correct?
		aa_flush(ascii_context);
	}


	// Thanks for lending us your buffer, you may have it back again:
        if (-1 == ioctl (fd, VIDIOC_QBUF, &buffer)) {
                perror ("VIDIOC_QBUF");
                exit (EXIT_FAILURE);
        }	
//	sleep(0.1);
}



// If we've finished, turn off streaming
if(-1 == ioctl(fd, VIDIOC_STREAMOFF, &buftype)) {
	perror("VIDIOC_STREAMOFF");
	exit(EXIT_FAILURE);
}


/* Cleanup. */

for (i = 0; i < reqbuf.count; i++)
        munmap (buffers[i].start, buffers[i].length);

aa_close(ascii_context);
free(grey);
if(fd>0) close(fd);
}// end main()


//Signal handling:
void quitproc(int Sig){
	printf("interrupt caught, exiting.\n");
	// turn off streaming
	if(-1 == ioctl(fd, VIDIOC_STREAMOFF, &buftype)) {
		perror("VIDIOC_STREAMOFF");
		exit(EXIT_FAILURE);
	}
	exit(0);
}


